"""
author: Andrea Pollastro - apollastro@lbl.gov
date:   04/19/2023
Python version: 3.10.8
"""
from datetime import datetime
import pytz

def to_UTC(datetime_obj: datetime, tznaive: bool) -> datetime:
        """ Converts datetime object to UTC datetime object.

        Params:
        - datetime_obj, datetime.datetime: datetime object - if tznaive, it will be 
                                           considered as US/Pacific tz (without 
                                           affecting the timestamp)
        - tznaive, bool: returns a datetime object tznaive or not

        Returns:
        - datetime_obj, datetime.datetime: datetime input converted to UTC time
        """
        if datetime_obj.tzinfo is None:
            datetime_obj.replace(tzinfo=pytz.timezone('US/Pacific'))
        datetime_obj = datetime_obj.astimezone(pytz.utc)
        if tznaive:
            datetime_obj = datetime_obj.replace(tzinfo=None)

        return datetime_obj

def to_PST(datetime_obj: datetime, tznaive: bool) -> datetime:
        """ Converts datetime object to PST datetime object.

        Params:
        - datetime_obj, datetime.datetime: datetime object - if tznaive, it will be 
                                           considered as UTC tz (without 
                                           affecting the timestamp)
        - tznaive, bool: returns a datetime object tznaive or not

        Returns:
        - datetime_obj, datetime.datetime: datetime input converted to PST time
        """
        if datetime_obj.tzinfo is None:
            datetime_obj.replace(tzinfo=pytz.utc)
        datetime_obj = datetime_obj.astimezone(pytz.timezone('US/Pacific'))
        if tznaive:
            datetime_obj = datetime_obj.replace(tzinfo=None)

        return datetime_obj


from typing import List
from ..pv import PV
from tqdm import tqdm
import pandas as pd

class DataPreprocesser:
    """
    This class preprocess (raw) data downloaded from the archiver and stored into PV objects.

    Methods (see function's docs for more info):
    - clean_data: it fills raw data with missing values according to their archiving policy. 
    - match_data: it matches data for a given list of PVs.
    """
    SEP = '==================================='
    
    def clean_data(self, pv: PV, precision: int) -> PV:
        """
        This function imputes missing values in the raw_data. Imputed data will be stored ad
        clean_data into the pv object. 
        
        params:
        - pv, PV: pv data
        - precision, int: data rate precision (ms)

        returns:
        - pv, PV: imputed pv.
        """
        pv.clean_data = pv.raw_data.drop(columns=['severity', 'status'])
        pv.clean_data = pv.clean_data.resample(f'{precision}ms').ffill()
        pv.clean_data.index = pv.clean_data.index.strftime('%Y-%m-%d %H:%M:%S.%f') # type: ignore
        pv.clean_data['val'].iloc[0] = pv.clean_data['val'].iloc[1] # first value is always NaN

        return pv
    
    def match_data(self, pv_list: List[PV], precision: int, verbose: bool = False) -> pd.DataFrame:
        """
        This function matches data listed in the list pv_list according to their timestamps.
        Using the precision param, you can define the timestamp precision in ms to upsample/downsample the series.

        params:
        - pv_list, List[PV]: list of PVs.
        - precision, int: data rate precision (ms)
        - strategy, str: reduction strategy. Values admitted: 'highest', 'lowest'.
        - verbose, bool (default, False): verbose

        returns:
        - matched_data, pd.DataFrame: matched data. Columns: [datetime(index), [PVs]]
        """
        pbar = tqdm(pv_list)
        for pv in pbar:
            pbar.set_description(f"Cleaning data PV {pv.name}")
            pv = self.clean_data(pv, precision)
        
        def extract_data(pv: PV) -> pd.DataFrame:
            data: pd.DataFrame = pv.clean_data
            data = data.rename(columns={'val': pv.name})
            return data
        
        # match data on secs and nanos
        matched_data: pd.DataFrame = extract_data(pv_list[0])
        for pv in pv_list[1:]:
            matched_data = matched_data.merge(extract_data(pv), on=['datetime'])
        
        if verbose:
            print(DataPreprocesser.SEP)
            print(f'PV matched: {[pv_name for pv_name in matched_data.columns]}')
            print(f'First timestamp: {matched_data.index[0]}')
            print(f'Last timestamp: {matched_data.index[-1]}')
            print(DataPreprocesser.SEP)

        return matched_data


from enum import Enum

class ArchivingPolicy(Enum):
    VERYSLOW = 0
    SLOW = 1
    MEDIUM = 2
    FAST = 3
    VERYFAST = 4

    def __lt__(self, other):
        if self.__class__ is other.__class__:
            return self.value < other.value
        return NotImplemented


from .archiving_policy import ArchivingPolicy
import pandas as pd

class PV:
    """
    This class represents the PV entity.
    
    Properties:
    - name, str: pv name.
    - raw_data, pd.DataFrame: raw data downloaded from the archiver.
    - properties, pd.DataFrame: pv properties as reported into the archiver.
    - clean_data, pd.DataFrame: cleaned data (with missing values imputed).
    - first_timestamp: pd.Timestamp: first timestamp of the timespan covered by the data.
    - last_timestamp: pd.Timestamp: last timestamp of the timespan covered by the data.
    - archiving_policy, ArchivingPolicy: archiving policy as reported on the archiver.
    Methods:
    - getter of the attributes
    - setter only of attributes clean_data and archiving_policy. It is suggested to don't
      change this values outside the other classes of the same package.
    """
    def __init__(self, name: str, raw_data: pd.DataFrame, properties: pd.DataFrame):
        self.__name: str = name
        self.__raw_data: pd.DataFrame = raw_data
        self.__properties: pd.DataFrame = properties
        self.__clean_data: pd.DataFrame = None  # type: ignore
        self.__first_timestamp: pd.Timestamp = raw_data.index[0] # type: ignore
        self.__last_timestamp: pd.Timestamp = raw_data.index[-1] # type: ignore

        def __extract_archiving_policy(pv_properties: pd.DataFrame) -> ArchivingPolicy:
            """
            Private function, it returns the archiving policy stored in the pv properties 
            in input as ArchivingPolicy object.

            params:
            - pv_properties, pd.DataFrame:

            returns:
            - archiving_policy, ArchivingPolicy: archiving policy of the pv into the archiver
            """
            ap_str = pv_properties[pv_properties['name'] == 'archive']['value'].values[0]
            ap_str = ap_str.lower().strip().replace('controlled', '')
            ap: ArchivingPolicy = ArchivingPolicy.FAST
            match ap_str:
                case 'veryfast':
                    ap = ArchivingPolicy.VERYFAST
                case 'fast':
                    ap = ArchivingPolicy.FAST
                case 'medium':
                    ap = ArchivingPolicy.MEDIUM
                case 'slow':
                    ap = ArchivingPolicy.SLOW
                case 'veryslow':
                    ap = ArchivingPolicy.VERYSLOW
            return ap
        self.__archiving_policy: ArchivingPolicy = __extract_archiving_policy(properties)

    @property
    def archiving_policy(self) -> ArchivingPolicy:
        return self.__archiving_policy
    
    @archiving_policy.setter
    def archiving_policy(self, archiving_policy) -> None:
        self.__archiving_policy = archiving_policy

    @property
    def first_timestamp(self) -> pd.Timestamp:
        return self.__first_timestamp
    
    @property
    def last_timestamp(self) -> pd.Timestamp:
        return self.__last_timestamp
    
    @property
    def name(self) -> str:
        return self.__name

    @property
    def raw_data(self) -> pd.DataFrame:
        return self.__raw_data
    
    @property
    def properties(self) -> pd.DataFrame:
        return self.__properties
    
    @property
    def clean_data(self) -> pd.DataFrame:
        return self.__clean_data
    
    @clean_data.setter
    def clean_data(self, clean_data: pd.DataFrame) -> None:
        self.__clean_data = clean_data

    def __repr__(self):
        return f'name:{self.name}, first timestamp:{self.first_timestamp}, last timestamp:{self.last_timestamp}, archiving policy:{self.archiving_policy}'


from .services.data_downloader import DataDownloader
from .services.data_preprocesser import DataPreprocesser
from .pv import PV
from datetime import datetime
from typing import List
from tqdm import tqdm
import pandas as pd
import warnings

class ArchiverClient:
    """
    This class manages the interaction between the users and the archiver.

    Properties (private):
    - __data_downloader: DataDownloader object
    - __data_preprocesser: DataPreprocesser object

    Methods (see function's docs for more info):
    - __match_data: it matches data for a given list of PVs.
    - download_data: it downloads data from a URL.
    - match_data: it matches data for a given list of PVs (string or PV objects).
    """
    def __init__(self, archiver_url: str = 'https://controls-web.als.lbl.gov'):
        self.__data_downloader = DataDownloader(archiver_url=archiver_url)
        self.__data_preprocesser = DataPreprocesser()

    def download_data(self, pv_name: str,
                            precision: int,
                            start: datetime, 
                            end: datetime,
                            verbose: bool = False) -> PV:
        """
        This function data downloaded from the archiver.

        params:
        - pv_name, str: pv name.
        - precision, int: data rate precision (ms)
        - start, datetime.datetime: start datetime in PST.
        - end, datetime.datetime: end datetime in PST.
        - verbose, bool (default, False): verbose level.

        returns:
        - pv, PV: pv data as PV object.
        """
        pv: PV =  self.__data_downloader.download_data(
            pv_name, start, end, verbose
        )
        pv = self.__data_preprocesser.clean_data(pv, precision)
        return pv
    
    def __match_data(self, pv_list: List[PV],
                         precision: int,
                         verbose: bool = True) -> pd.DataFrame:
        """
        Private function for calling match_data on list ov PV objects.

        params:
        - pv_list, List[PV]|List[str]: list of PVs.
        - precision, int: data rate precision (ms)
        - strategy, str: reduction strategy. Values admitted: 'highest', 'lowest'.
        - verbose, bool (default, False): verbose

        returns:
        - matched_data, pd.DataFrame: matched data. Columns: [secs, nanos, [PVs]]
        """
        return self.__data_preprocesser.match_data(pv_list, precision, verbose)

    def match_data(self, pv_list: List[str], # type: ignore
                         precision: int,
                         start: datetime, 
                         end: datetime, 
                         verbose: int = 0) -> pd.DataFrame:
        """
        This function matches PVs defined in pv_list.

        params:
        - pv_list, List[PV]|List[str]: list of PVs.
        - precision, int: data rate precision (ms).
        - start, datetime.datetime: start datetime in PST.
        - end, datetime.datetime: end datetime in PST.
        - verbose, bool (default, True): verbose level, 0 to deactivate, 1 only for merge, 2 for merge and download.

        returns:
        - matched_data, pd.DataFrame: matched data. Columns: [secs, nanos, [PVs]]
        """
        assert all(filter(lambda x: isinstance(x, str), pv_list)), 'please use only lists of strings.'
        verbose_download = verbose == 2
        verbose_match = verbose > 0
        pv_list_obj: List[PV] = []
        pbar = tqdm(pv_list)
        for pv in pbar:
            pbar.set_description(f"Downloading PV {pv}")
            try:
                pv = self.download_data(pv, precision, start, end, verbose_download)
                pv_list_obj.append(pv) # type: ignore
            except:
                warnings.warn(f'An error occurred while fetching {pv} data. PV skipped.')
        return self.__match_data(pv_list_obj, precision, verbose_match) # type: ignore


from .archiver_client import ArchiverClient
from .archiving_policy import ArchivingPolicy
from .pv import PV

__all__ = [
    'archiver_client',
    'archiving_policy',
    'pv',
]

__version__ = '1.0.0'
__author__ = 'Andrea Pollastro'
__author_email__ = 'apollastro@lbl.gov'


from ..utils import to_UTC
from urllib3 import PoolManager, disable_warnings
disable_warnings()
from datetime import datetime
from typing import List
from ..pv import PV
import os
import json
import warnings
import numpy as np
import pandas as pd
from urllib.parse import urlparse


class DataDownloader:
    """
    This class manages the download and the interactions with the archiver server.

    Properties (private):
    - ARCHIVER_URL, str: URL to archiver.
    - DATA_JSON, str: URL portion (to combine with ARCHIVER_URL) to download json data.
    - CHANNEL_FINDER, str: URL portion (to combine with ARCHIVER_URL) to download pv properties.
    - URL portion (to combine with ARCHIVER_URL) to download json data.

    Methods (see function's docs for more info):
    - __ping_archiver: it verifies if the archiver is reachable.
    - __http_request: it does http requests to archiver.
    - __download_data: it downloads data from a URL.
    - __filter_data: it filters data in a given timespan.
    - __pv_properties: it downloads the pv properties.
    - download_data: it downloads data from a URL.
    """
    def __init__(self, archiver_url: str = None):
        self.__ARCHIVER_URL = archiver_url
        self.__DATA_JSON: str = '/archappl_retrieve/data/getData.json?'
        self.__CHANNEL_FINDER: str = '/ChannelFinder/resources/channels?'
        self.__SEP = '==================================='

        is_reachable = self.__ping_archiver()
        if is_reachable is False:
            raise ConnectionError("Archiver server is unreachable. Hint: activate the ALS VPN and restart the application.")
        print('Archiver server is reachable via ping.')
    
    @property
    def archiver_url(self) -> str:
        return self.__ARCHIVER_URL
    
    def __ping_archiver(self) -> bool:
        """This function verifies if the archiver's server is reachable via ping. It is called only in the constructor"""
        print(self.__SEP)
        print("Verifying the reachability of the archiver's server...")
        parsed_url = urlparse(self.__ARCHIVER_URL)
        hostname = parsed_url.netloc.split(':')[0]
        exit_status: int = os.system(f"ping -c 1 {hostname}")
        print(self.__SEP)
        return exit_status == 0

    def __http_request(self, url: str) -> List:
        """
        http request to archiver server.

        params:
        - url, str: url built on request type (ask to Tynan for more info).

        returns:
        - json_data, list: data retrieved from http request.
        """
        http: PoolManager = PoolManager(cert_reqs='CERT_NONE')
        response = http.request('GET', url)
        data: bytes = response.data
        json_data: list = json.loads(data)        
        return json_data

    def __download_data(self, pv_name: str, 
                            start: datetime, 
                            end: datetime,
                            verbose: bool = False) -> pd.DataFrame:
        """
        This function returns the raw data from the archiver.

        params:
        - pv_name, str: pv_name name
        - start, datetime.datetime: start datetime in PST.
        - end, datetime.datetime: end datetime in PST.
        - verbose, bool (default, False): verbose level (default, True).

        returns:
        - data, pd.DataFrame: raw data.
        """
        if verbose: print(f"Downloading data from {start} to {end}")
        
        # archiver's http requests want dates in UTC
        start = to_UTC(start, tznaive=True)
        end = to_UTC(end, tznaive=True)

        # building URL
        url_components: list = [
            self.__ARCHIVER_URL,
            self.__DATA_JSON,
            f'pv={pv_name}&',
            f'from={start.isoformat()}Z&'
            f'to={end.isoformat()}Z&',
            'fetchLatestMetadata=true',
        ]
        url: str = "".join(url_components)
        
        # downloading data
        data: dict = self.__http_request(url)[0]

        # organizing data as dictionary
        data_keys: list = list(data['data'][0].keys())
        if 'fields' in data_keys: 
            data_keys.remove('fields')
        df: pd.DataFrame = pd.DataFrame({k: np.array([d[k] for d in data['data']]) for k in data_keys})
                
        return df
    
    def __filter_data(self, df: pd.DataFrame, start_ts: int, end_ts: int) -> pd.DataFrame:
        """
        This function returns the dataframe filtered by a timespan defined by the timestamps in input

        params:
        - df, pd.DataFrame: pv data.
        - start_ts, int: first timestamp.
        - end_ts, int: last timestamp.

        returns:
        - df, pd.DataFrame: filterd df.
        """
        if start_ts not in df['secs'].values:
            record: pd.DataFrame = df[df['secs'] < start_ts].tail(1)
            record['secs'] = start_ts
            idx: int = int(record.index[0]) + 1 # type: ignore
            df = pd.concat([df.iloc[:idx], record, df.iloc[idx:]]).reset_index(drop=True)
        if end_ts not in df['secs'].values:
            record: pd.DataFrame = df[df['secs'] < end_ts].tail(1)
            record['secs'] = end_ts
            idx: int = (record.index[0]) + 1 # type: ignore
            df = pd.concat([df.iloc[:idx], record, df.iloc[idx:]]).reset_index(drop=True)
        left_mask: pd.Series = df['secs'] >= start_ts
        right_mask: pd.Series = df['secs'] <= end_ts
        df = df.loc[left_mask & right_mask, :]
        return df

    def __pv_properties(self, pv: str) -> pd.DataFrame:
        """
        This function returns the properties of the PV in input as pandas DataFrame.

        params:
        - pv, str: PV name.

        returns:
        - properties, pd.DataFrame: dataframe containing properties of the PV.
        """
        url_components: list = [
            self.__ARCHIVER_URL,
            self.__CHANNEL_FINDER,
            f'~name={pv}'
        ]
        url = "".join(url_components)
        properties = self.__http_request(url)
        if len(properties) == 0: raise ValueError(f'PV {pv} not found.')
        if len(properties) == 0: warnings.warn('Multiple PVs retrieved, only the first one will be returned.')

        return pd.DataFrame(properties[0]['properties'])

    def download_data(self, pv_name: str, 
                            start: datetime, 
                            end: datetime,
                            verbose: bool = False) -> PV:
        """
        This function returns raw data downloaded from the archiver.

        params:
        - pv_name, str: pv name.
        - start, datetime.datetime: start datetime in PST.
        - end, datetime.datetime: end datetime in PST.
        - verbose, bool (default, False): verbose level.

        returns:
        - pv, PV: pv data as PV object.
        """
        start_ts: int = int(start.timestamp())
        end_ts: int = int(end.timestamp())
        assert start_ts < end_ts, 'Invalid: start date must be before end date'
        df: pd.DataFrame
        
        if verbose: 
            print(self.__SEP)
            print(f"Downloading data for pv {pv_name}")
        pv_properties = self.__pv_properties(pv_name)

        # Data download
        df = self.__download_data(pv_name, start, end, verbose=verbose)

        # What usually happens when we download data from the archiver for the first time is that 
        # (i) the first timestamp available is always lower than start 
        # (ii) the last timestamp too. 
        # For this reason, we introduce two timestamps having the values of our query equal
        # to the last valid values before. In this way, we can return the exact timespan
        df = self.__filter_data(df, start_ts, end_ts)

        # creating datetime index
        datetime_series = pd.to_datetime(df['secs'] + df['nanos']*1e-9, unit='s', utc=True).dt.tz_convert('US/Pacific')
        df = df.set_index(pd.DatetimeIndex(datetime_series, name='datetime'))
        df = df.drop(columns=['secs', 'nanos'])

        if verbose:
            print(f'First timestamp: {df.index[0]}')
            print(f'Last timestamp: {df.index[-1]}')
            print(self.__SEP)

        # Create PV object
        pv: PV = PV(name=pv_name, raw_data=df, properties=pv_properties)
            
        return pv



from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="als-archiver-client",
    version="1.0.2",
    author="Andrea Pollastro",
    author_email="apollastro@lbl.gov",
    description="A Python package for interacting with the ALS archiver.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/andrea-pollastro/als-archiver-client",
    packages=find_packages(),
    install_requires=[
        "numpy>=1.26.3",
        "pandas>=2.1.4",
        "python-dateutil>=2.8.2",
        "pytz>=2023.3",
        "six>=1.16.0",
        "tzdata>=2023.3",
        "urllib3>=2.0.0",
        "tqdm>=4.65.0",
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    python_requires='>=3.10',
    keywords="epics, archiver, data-mining, engineering, research, particle-accelerator",
)
